#include "defs.h"

#include "communicate.h"


/*
 * This is part of JEliza 2.3.
 * Copyright 2006 by Tobias Schulz
 * WWW: http://jeliza.ch.to/
 *
 * JEliza is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later
 * version.
 *
 * JEliza is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU GPL
 * along with JEliza (file gpl.txt) ; if not, write
 * to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

typedef int COMMUNICATOR_TYPE;
COMMUNICATOR_TYPE SERVER = 1;
COMMUNICATOR_TYPE CLIENT = 2;

COMMUNICATOR_TYPE communicator_type;

string host = "127.0.0.1";
unsigned short port = 5173;

vector<jeliza::string> to_send;
vector<boost::shared_ptr<tcp::iostream> > streams;
bool ok_to_write = false;
bool ready_to_write = false;


const int max_length = 1024;
typedef boost::shared_ptr<tcp::socket> socket_ptr;
asio::io_service io_service;

void jeliza::msleep (unsigned long msecs) {
    asio::deadline_timer timer(io_service, boost::posix_time::milliseconds(msecs));
    timer.wait();
}


void session_stream_read(boost::shared_ptr<tcp::iostream> stream) {
    string line;
    for (;;) {
        asio::deadline_timer timer(io_service, boost::posix_time::milliseconds(50));
        try {
            getline(*stream, line);

            while (line.size() > 0) {
                jeliza::comm_new(jeliza::string(line));

                getline(*stream, line);
            }

            timer.wait();
        }
        catch (std::exception& e) {
//            std::cerr << 1 << "Exception in thread: " << e.what() << "\n";
        }
    }
}

unsigned long in_schlange = 0;

void comm_write(std::string str) {
    for (unsigned int x = 0; x < streams.size(); x++) {
        try {
            ++in_schlange;
            do {
                ready_to_write = false;
                (*(streams[x])) << str << endl;
                ready_to_write = true;
            } while (!ready_to_write);
            --in_schlange;
        }
        catch (std::exception& e) {
            std::cerr << 3 << "Exception in thread: " << e.what() << "\n";
        }
    }
}
void flush_streams() {
    while (true) {
        for (unsigned int x = 0; x < streams.size(); x++) {
            try {
                (*(streams[x])).flush();
            }
            catch (std::exception& e) {
                std::cerr << 3 << "Exception in thread: " << e.what() << "\n";
            }
        }

        asio::deadline_timer timer(io_service, boost::posix_time::milliseconds(500));
        timer.wait();
    }
}

void io_service_run() {
    io_service.run();
}

void send_hello() {
    asio::io_service io_service;
    asio::deadline_timer timer(io_service, boost::posix_time::milliseconds(2000));
    timer.wait();

    jeliza::comm_send("HELLO:" + KI_NAME);
    jeliza::send_name();
}

void server() {
    asio::thread t_ios(io_service_run);

    try {
        tcp::acceptor a(io_service, tcp::endpoint(tcp::v4(), port));
        for (;;) {
            try {
                boost::shared_ptr<tcp::iostream> stream(new tcp::iostream);
                a.accept(*stream->rdbuf());
                asio::ip::tcp::no_delay option(true);
                stream->rdbuf()->set_option(option);

                cout << ":: Neuer Client" << endl;

                asio::thread t_r(boost::bind(session_stream_read, stream));
                asio::thread t_f(flush_streams);
                cout << ":: Threads started (" << streams.size() << ")" << endl;
                streams.push_back(stream);
                ok_to_write = true;
                ready_to_write = true;

                asio::thread t_h(send_hello);
            }
            catch (std::exception& e) {
                std::cerr << 5.5 << "Exception in thread: " << e.what() << "\n";
            }
        }
    }
    catch (std::exception& e) {
        std::cerr << 5 << "Exception in thread: " << e.what() << "\n";
    }
}

void client(std::string host, bool verbose = true, bool wait = false) {
    if (wait) {
        asio::deadline_timer timer(io_service, boost::posix_time::milliseconds(2000));
        timer.wait();
    }

    asio::thread t_ios(io_service_run);

    try {
        tcp::endpoint endp(asio::ip::address_v4::from_string(host.c_str()), (unsigned short)(port));
        boost::shared_ptr<tcp::iostream> stream(new tcp::iostream);
        stream->rdbuf()->connect(endp);
//        asio::ip::tcp::no_delay option(true);
//        stream->rdbuf()->set_option(option);

        if (verbose) {
            cout << ":: Connected" << endl;
        }

        asio::thread t_r(boost::bind(session_stream_read, stream));
        asio::thread t_f(flush_streams);
        if (verbose) {
            cout << ":: Threads started" << endl;
        }
        streams.push_back(stream);
        ok_to_write = true;
        ready_to_write = true;
    }
    catch (std::exception& e) {
        std::cerr << 4 << "Exception in thread: " << e.what() << "\n";
    }
}



void jeliza::comm_init_server() {
    communicator_type = SERVER;

    asio::thread t(server);
}


void jeliza::comm_init_client(std::string host, bool verbose, bool wait = false) {
    communicator_type = CLIENT;

    asio::thread t(boost::bind(client, host, verbose, wait));
}


void jeliza::comm_send(jeliza::string s) {
    if (ok_to_write) {
        if (s.contains("ĹOG")) {
            cout << "--" << s.ref() << endl;
        }
        comm_write(s.ref());
    }
}

void jeliza::comm_send(std::string s) {
    comm_send(jeliza::string(s));
}

void jeliza::send_name() {
    jeliza::comm_send("NAME:" + KI_NAME);
    jeliza::comm_send("JELIZA_FULL_VERSION:" + JELIZA_FULL_VERSION);
}
